<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Export;

use Chill\MainBundle\Entity\Address;
use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Export\FormatterInterface;
use Chill\MainBundle\Export\GroupedExportInterface;
use Chill\MainBundle\Export\Helper\DateTimeHelper;
use Chill\MainBundle\Export\Helper\ExportAddressHelper;
use Chill\MainBundle\Export\Helper\UserHelper;
use Chill\MainBundle\Export\ListInterface;
use Chill\MainBundle\Form\Type\PickRollingDateType;
use Chill\MainBundle\Service\RollingDate\RollingDateConverterInterface;
use Chill\MainBundle\Templating\TranslatableStringHelperInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\AccompanyingPeriodParticipation;
use Chill\PersonBundle\Entity\Household\PersonHouseholdAddress;
use Chill\PersonBundle\Entity\Person\PersonCenterHistory;
use Chill\PersonBundle\Entity\SocialWork\SocialIssue;
use Chill\PersonBundle\Export\Declarations;
use Chill\PersonBundle\Repository\PersonRepository;
use Chill\PersonBundle\Repository\SocialWork\SocialIssueRepository;
use Chill\PersonBundle\Security\Authorization\PersonVoter;
use Chill\PersonBundle\Templating\Entity\PersonRenderInterface;
use Chill\PersonBundle\Templating\Entity\SocialIssueRender;
use Chill\ThirdPartyBundle\Repository\ThirdPartyRepository;
use Chill\ThirdPartyBundle\Templating\Entity\ThirdPartyRender;
use DateTimeImmutable;
use Doctrine\ORM\AbstractQuery;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Query\Expr\Join;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormBuilderInterface;
use function strlen;

class ListAccompanyingPeriod implements ListInterface, GroupedExportInterface
{
    private const FIELDS = [
        'id',
        'step',
        'stepSince',
        'openingDate',
        'closingDate',
        'referrer',
        'referrerSince',
        'administrativeLocation',
        'locationIsPerson',
        'locationIsTemp',
        'locationPersonName',
        'locationPersonId',
        'origin',
        'closingMotive',
        'confidential',
        'emergency',
        'intensity',
        'job',
        'isRequestorPerson',
        'isRequestorThirdParty',
        'requestorPerson',
        'requestorPersonId',
        'requestorThirdParty',
        'requestorThirdPartyId',
        'scopes',
        'socialIssues',
        'createdAt',
        'createdBy',
        'updatedAt',
        'updatedBy',
    ];

    private ExportAddressHelper $addressHelper;

    private DateTimeHelper $dateTimeHelper;

    private EntityManagerInterface $entityManager;

    private PersonRenderInterface $personRender;

    private PersonRepository $personRepository;

    private RollingDateConverterInterface $rollingDateConverter;

    private SocialIssueRender $socialIssueRender;

    private SocialIssueRepository $socialIssueRepository;

    private ThirdPartyRender $thirdPartyRender;

    private ThirdPartyRepository $thirdPartyRepository;

    private TranslatableStringHelperInterface $translatableStringHelper;

    private UserHelper $userHelper;

    public function __construct(
        ExportAddressHelper $addressHelper,
        DateTimeHelper $dateTimeHelper,
        EntityManagerInterface $entityManager,
        PersonRenderInterface $personRender,
        PersonRepository $personRepository,
        ThirdPartyRepository $thirdPartyRepository,
        ThirdPartyRender $thirdPartyRender,
        SocialIssueRepository $socialIssueRepository,
        SocialIssueRender $socialIssueRender,
        TranslatableStringHelperInterface $translatableStringHelper,
        RollingDateConverterInterface $rollingDateConverter,
        UserHelper $userHelper
    ) {
        $this->addressHelper = $addressHelper;
        $this->dateTimeHelper = $dateTimeHelper;
        $this->entityManager = $entityManager;
        $this->personRender = $personRender;
        $this->personRepository = $personRepository;
        $this->socialIssueRender = $socialIssueRender;
        $this->socialIssueRepository = $socialIssueRepository;
        $this->thirdPartyRender = $thirdPartyRender;
        $this->thirdPartyRepository = $thirdPartyRepository;
        $this->translatableStringHelper = $translatableStringHelper;
        $this->rollingDateConverter = $rollingDateConverter;
        $this->userHelper = $userHelper;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder
            ->add('calc_date', PickRollingDateType::class, [
                'label' => 'export.list.acp.Date of calculation for associated elements',
                'help' => 'export.list.acp.The associated referree, localisation, and other elements will be valid at this date',
                'required' => true,
            ]);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_LIST];
    }

    public function getDescription()
    {
        return 'export.list.acp.Generate a list of accompanying periods, filtered on different parameters.';
    }

    public function getGroup(): string
    {
        return 'Exports of accompanying courses';
    }

    public function getLabels($key, array $values, $data)
    {
        if (substr($key, 0, strlen('address_fields')) === 'address_fields') {
            return $this->addressHelper->getLabel($key, $values, $data, 'address_fields');
        }

        switch ($key) {
            case 'stepSince':
            case 'openingDate':
            case 'closingDate':
            case 'referrerSince':
            case 'createdAt':
            case 'updatedAt':
                return $this->dateTimeHelper->getLabel('export.list.acp.' . $key);

            case 'origin':
            case 'closingMotive':
            case 'job':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acp.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return $this->translatableStringHelper->localize(json_decode($value, true));
                };

            case 'locationPersonName':
            case 'requestorPerson':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acp.' . $key;
                    }

                    if (null === $value || null === $person = $this->personRepository->find($value)) {
                        return '';
                    }

                    return $this->personRender->renderString($person, []);
                };

            case 'requestorThirdParty':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acp.' . $key;
                    }

                    if (null === $value || null === $thirdparty = $this->thirdPartyRepository->find($value)) {
                        return '';
                    }

                    return $this->thirdPartyRender->renderString($thirdparty, []);
                };

            case 'scopes':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acp.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return implode(
                        '|',
                        array_map(
                            fn ($s) => $this->translatableStringHelper->localize($s),
                            json_decode($value, true)
                        )
                    );
                };

            case 'socialIssues':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acp.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return implode(
                        '|',
                        array_map(
                            fn ($s) => $this->socialIssueRender->renderString($this->socialIssueRepository->find($s), []),
                            json_decode($value, true)
                        )
                    );
                };

            default:
                return static function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acp.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return $value;
                };
        }
    }

    public function getQueryKeys($data)
    {
        return array_merge(
            self::FIELDS,
            $this->addressHelper->getKeys(ExportAddressHelper::F_ALL, 'address_fields')
        );
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(AbstractQuery::HYDRATE_SCALAR);
    }

    public function getTitle()
    {
        return 'export.list.acp.List of accompanying periods';
    }

    public function getType()
    {
        return Declarations::PERSON_TYPE;
    }

    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(static function ($el) {
            return $el['center'];
        }, $acl);

        $qb = $this->entityManager->createQueryBuilder();

        $qb
            ->from(AccompanyingPeriod::class, 'acp')
            ->andWhere('acp.step != :list_acp_step')
            ->andWhere(
                $qb->expr()->exists(
                    'SELECT 1 FROM ' . AccompanyingPeriodParticipation::class . ' acl_count_part
                    JOIN ' . PersonCenterHistory::class . ' acl_count_person_history WITH IDENTITY(acl_count_person_history.person) = IDENTITY(acl_count_part.person)
                    WHERE acl_count_part.accompanyingPeriod = acp.id AND acl_count_person_history.center IN (:authorized_centers)
                    '
                )
            )
            ->setParameter('list_acp_step', AccompanyingPeriod::STEP_DRAFT)
            ->setParameter('authorized_centers', $centers);

        $this->addSelectClauses($qb, $this->rollingDateConverter->convert($data['calc_date']));

        return $qb;
    }

    public function requiredRole(): string
    {
        return PersonVoter::LISTS;
    }

    public function supportsModifiers()
    {
        return [
            Declarations::ACP_TYPE,
        ];
    }

    private function addSelectClauses(QueryBuilder $qb, DateTimeImmutable $calcDate): void
    {
        // add the regular fields
        foreach (['id', 'openingDate', 'closingDate', 'confidential', 'emergency', 'intensity', 'createdAt', 'updatedAt'] as $field) {
            $qb->addSelect(sprintf('acp.%s AS %s', $field, $field));
        }

        // add the field which are simple association
        foreach (['origin' => 'label', 'closingMotive' => 'name', 'job' => 'label', 'createdBy' => 'label', 'updatedBy' => 'label', 'administrativeLocation' => 'name'] as $entity => $field) {
            $qb
                ->leftJoin(sprintf('acp.%s', $entity), "{$entity}_t")
                ->addSelect(sprintf('%s_t.%s AS %s', $entity, $field, $entity));
        }

        // step at date
        $qb
            ->addSelect('stepHistory.step AS step')
            ->addSelect('stepHistory.startDate AS stepSince')
            ->leftJoin('acp.stepHistories', 'stepHistory')
            ->andWhere(
                $qb->expr()->andX(
                    $qb->expr()->lte('stepHistory.startDate', ':calcDate'),
                    $qb->expr()->orX($qb->expr()->isNull('stepHistory.endDate'), $qb->expr()->gt('stepHistory.endDate', ':calcDate'))
                )
            );

        // referree at date
        $qb
            ->addSelect('referrer_t.label AS referrer')
            ->addSelect('userHistory.startDate AS referrerSince')
            ->leftJoin('acp.userHistories', 'userHistory')
            ->leftJoin('userHistory.user', 'referrer_t')
            ->andWhere(
                $qb->expr()->orX(
                    $qb->expr()->isNull('userHistory'),
                    $qb->expr()->andX(
                        $qb->expr()->lte('userHistory.startDate', ':calcDate'),
                        $qb->expr()->orX($qb->expr()->isNull('userHistory.endDate'), $qb->expr()->gt('userHistory.endDate', ':calcDate'))
                    )
                )
            );

        // location of the acp
        $qb
            ->addSelect('CASE WHEN locationHistory.personLocation IS NOT NULL THEN 1 ELSE 0 END AS locationIsPerson')
            ->addSelect('CASE WHEN locationHistory.personLocation IS NOT NULL THEN 0 ELSE 1 END AS locationIsTemp')
            ->addSelect('IDENTITY(locationHistory.personLocation) AS locationPersonName')
            ->addSelect('IDENTITY(locationHistory.personLocation) AS locationPersonId')
            ->leftJoin('acp.locationHistories', 'locationHistory')
            ->andWhere(
                $qb->expr()->orX(
                    $qb->expr()->isNull('locationHistory'),
                    $qb->expr()->andX(
                        $qb->expr()->lte('locationHistory.startDate', ':calcDate'),
                        $qb->expr()->orX($qb->expr()->isNull('locationHistory.endDate'), $qb->expr()->gt('locationHistory.endDate', ':calcDate'))
                    )
                )
            )
            ->leftJoin(PersonHouseholdAddress::class, 'personAddress', Join::WITH, 'locationHistory.personLocation = personAddress.person')
            ->andWhere(
                $qb->expr()->orX(
                    $qb->expr()->isNull('personAddress'),
                    $qb->expr()->andX(
                        $qb->expr()->lte('personAddress.validFrom', ':calcDate'),
                        $qb->expr()->orX($qb->expr()->isNull('personAddress.validTo'), $qb->expr()->gt('personAddress.validTo', ':calcDate'))
                    )
                )
            )
            ->leftJoin(Address::class, 'acp_address', Join::WITH, 'COALESCE(IDENTITY(locationHistory.addressLocation), IDENTITY(personAddress.address)) = acp_address.id');

        $this->addressHelper->addSelectClauses(ExportAddressHelper::F_ALL, $qb, 'acp_address', 'address_fields');

        // requestor
        $qb
            ->addSelect('CASE WHEN acp.requestorPerson IS NULL THEN 1 ELSE 0 END AS isRequestorPerson')
            ->addSelect('CASE WHEN acp.requestorPerson IS NULL THEN 0 ELSE 1 END AS isRequestorThirdParty')
            ->addSelect('IDENTITY(acp.requestorPerson) AS requestorPersonId')
            ->addSelect('IDENTITY(acp.requestorThirdParty) AS requestorThirdPartyId')
            ->addSelect('IDENTITY(acp.requestorPerson) AS requestorPerson')
            ->addSelect('IDENTITY(acp.requestorThirdParty) AS requestorThirdParty');

        $qb
            // scopes
            ->addSelect('(SELECT AGGREGATE(scope.name) FROM ' . Scope::class . ' scope WHERE scope MEMBER OF acp.scopes) AS scopes')
            // social issues
            ->addSelect('(SELECT AGGREGATE(socialIssue.id) FROM ' . SocialIssue::class . ' socialIssue WHERE socialIssue MEMBER OF acp.socialIssues) AS socialIssues');

        // add parameter
        $qb->setParameter('calcDate', $calcDate);
    }
}
