<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Export;

use Chill\MainBundle\Export\FormatterInterface;
use Chill\MainBundle\Export\GroupedExportInterface;
use Chill\MainBundle\Export\Helper\AggregateStringHelper;
use Chill\MainBundle\Export\Helper\DateTimeHelper;
use Chill\MainBundle\Export\Helper\TranslatableStringExportLabelHelper;
use Chill\MainBundle\Export\Helper\UserHelper;
use Chill\MainBundle\Export\ListInterface;
use Chill\MainBundle\Form\Type\PickRollingDateType;
use Chill\MainBundle\Service\RollingDate\RollingDate;
use Chill\MainBundle\Service\RollingDate\RollingDateConverterInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod\AccompanyingPeriodWork;
use Chill\PersonBundle\Entity\AccompanyingPeriod\AccompanyingPeriodWorkEvaluation;
use Chill\PersonBundle\Entity\AccompanyingPeriod\AccompanyingPeriodWorkGoal;
use Chill\PersonBundle\Entity\AccompanyingPeriod\UserHistory;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Entity\Person\PersonCenterHistory;
use Chill\PersonBundle\Entity\SocialWork\Evaluation;
use Chill\PersonBundle\Entity\SocialWork\Goal;
use Chill\PersonBundle\Entity\SocialWork\Result;
use Chill\PersonBundle\Export\Declarations;
use Chill\PersonBundle\Export\Helper\LabelPersonHelper;
use Chill\PersonBundle\Repository\SocialWork\SocialActionRepository;
use Chill\PersonBundle\Repository\SocialWork\SocialIssueRepository;
use Chill\PersonBundle\Security\Authorization\AccompanyingPeriodVoter;
use Chill\PersonBundle\Templating\Entity\SocialActionRender;
use Chill\PersonBundle\Templating\Entity\SocialIssueRender;
use Chill\ThirdPartyBundle\Entity\ThirdParty;
use Chill\ThirdPartyBundle\Export\Helper\LabelThirdPartyHelper;
use DateTimeImmutable;
use Doctrine\ORM\AbstractQuery;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormBuilderInterface;

class ListAccompanyingPeriodWork implements ListInterface, GroupedExportInterface
{
    private const FIELDS = [
        'id',
        'socialActionId',
        'socialAction',
        'socialIssue',
        'acp_id',
        'acp_user',
        'startDate',
        'endDate',
        'goalsId',
        'goalsTitle',
        'goalResultsId',
        'goalResultsTitle',
        'resultsId',
        'resultsTitle',
        'evaluationsId',
        'evaluationsTitle',
        'note',
        'personsId',
        'personsName',
        'thirdParties',
        'handlingThierParty',
        'referrers',
        'createdAt',
        'createdBy',
        'updatedAt',
        'updatedBy',
    ];

    private AggregateStringHelper $aggregateStringHelper;

    private DateTimeHelper $dateTimeHelper;

    private EntityManagerInterface $entityManager;

    private LabelPersonHelper $personHelper;

    private RollingDateConverterInterface $rollingDateConverter;

    private SocialActionRender $socialActionRender;

    private SocialActionRepository $socialActionRepository;

    private SocialIssueRender $socialIssueRender;

    private SocialIssueRepository $socialIssueRepository;

    private LabelThirdPartyHelper $thirdPartyHelper;

    private TranslatableStringExportLabelHelper $translatableStringExportLabelHelper;

    private UserHelper $userHelper;

    public function __construct(
        EntityManagerInterface $entityManager,
        DateTimeHelper $dateTimeHelper,
        UserHelper $userHelper,
        LabelPersonHelper $personHelper,
        LabelThirdPartyHelper $thirdPartyHelper,
        TranslatableStringExportLabelHelper $translatableStringExportLabelHelper,
        SocialIssueRender $socialIssueRender,
        SocialIssueRepository $socialIssueRepository,
        SocialActionRender $socialActionRender,
        RollingDateConverterInterface $rollingDateConverter,
        AggregateStringHelper $aggregateStringHelper,
        SocialActionRepository $socialActionRepository
    ) {
        $this->entityManager = $entityManager;
        $this->dateTimeHelper = $dateTimeHelper;
        $this->userHelper = $userHelper;
        $this->personHelper = $personHelper;
        $this->thirdPartyHelper = $thirdPartyHelper;
        $this->translatableStringExportLabelHelper = $translatableStringExportLabelHelper;
        $this->socialIssueRender = $socialIssueRender;
        $this->socialIssueRepository = $socialIssueRepository;
        $this->socialActionRender = $socialActionRender;
        $this->rollingDateConverter = $rollingDateConverter;
        $this->aggregateStringHelper = $aggregateStringHelper;
        $this->socialActionRepository = $socialActionRepository;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder
            ->add('calc_date', PickRollingDateType::class, [
                'label' => 'export.list.acpw.Date of calculation for associated elements',
                'help' => 'export.list.acpw.help_description',
                'required' => true,
                'data' => new RollingDate(RollingDate::T_TODAY),
            ]);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_LIST];
    }

    public function getDescription(): string
    {
        return 'export.list.acpw.List description';
    }

    public function getGroup(): string
    {
        return 'Exports of social work actions';
    }

    public function getLabels($key, array $values, $data)
    {
        switch ($key) {
            case 'startDate':
            case 'endDate':
            case 'createdAt':
            case 'updatedAt':
                return $this->dateTimeHelper->getLabel('export.list.acpw.' . $key);

            case 'socialAction':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acpw.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return $this->socialActionRender->renderString(
                        $this->socialActionRepository->find($value),
                        []
                    );
                };

            case 'socialIssue':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acpw.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return $this->socialIssueRender->renderString(
                        $this->socialIssueRepository->find($value),
                        []
                    );
                };

            case 'createdBy':
            case 'updatedBy':
            case 'acp_user':
                return $this->userHelper->getLabel($key, $values, 'export.list.acpw.' . $key);

            case 'referrers':
                //$date = $this->rollDateConverter->convert($data['calc_date'])->format('d/m/Y');
                return $this->userHelper->getLabel($key, $values, 'export.list.acpw.' . $key);

            case 'personsName':
                return $this->personHelper->getLabelMulti($key, $values, 'export.list.acpw.' . $key);

            case 'handlingThierParty':
                return $this->thirdPartyHelper->getLabel($key, $values, 'export.list.acpw.' . $key);

            case 'thirdParties':
                return $this->thirdPartyHelper->getLabelMulti($key, $values, 'export.list.acpw.' . $key);

            case 'personsId':
            case 'goalsId':
            case 'goalResultsId':
            case 'resultsId':
            case 'evaluationsId':
                return $this->aggregateStringHelper->getLabelMulti($key, $values, 'export.list.acpw.' . $key);

            case 'goalsTitle':
            case 'goalResultsTitle':
            case 'resultsTitle':
            case 'evaluationsTitle':
                return $this->translatableStringExportLabelHelper->getLabelMulti($key, $values, 'export.list.acpw.' . $key);

            default:
                return static function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.acpw.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return $value;
                };
        }
    }

    public function getQueryKeys($data)
    {
        return self::FIELDS;
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(AbstractQuery::HYDRATE_SCALAR);
    }

    public function getTitle(): string
    {
        return 'export.list.acpw.List of accompanying period works';
    }

    public function getType(): string
    {
        return Declarations::SOCIAL_WORK_ACTION_TYPE;
    }

    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(static function ($el) {
            return $el['center'];
        }, $acl);

        $qb = $this->entityManager->createQueryBuilder();

        $qb
            ->from(AccompanyingPeriodWork::class, 'acpw')
            ->distinct()
            ->select('acpw.id AS id')
            ->join('acpw.accompanyingPeriod', 'acp')
            ->join('acp.participations', 'acppart')
            ->join('acppart.person', 'person')
            // ignore participation which didn't last one day, at least
            ->andWhere('acppart.startDate != acppart.endDate OR acppart.endDate IS NULL')
            // get participants at the given date
            ->andWhere('acppart.startDate <= :calc_date AND (acppart.endDate > :calc_date OR acppart.endDate IS NULL)')
            ->andWhere(
                $qb->expr()->exists(
                    'SELECT 1 FROM ' . PersonCenterHistory::class . ' acl_count_person_history WHERE acl_count_person_history.person = person
                    AND acl_count_person_history.center IN (:authorized_centers)
                    '
                )
            )
            ->setParameter('authorized_centers', $centers)
            ->setParameter('calc_date', $this->rollingDateConverter->convert($data['calc_date']));

        $this->addSelectClauses($qb, $this->rollingDateConverter->convert($data['calc_date']));

        return $qb;
    }

    public function requiredRole(): string
    {
        return AccompanyingPeriodVoter::STATS;
    }

    public function supportsModifiers(): array
    {
        return [
            Declarations::SOCIAL_WORK_ACTION_TYPE,
            Declarations::ACP_TYPE,
            Declarations::PERSON_TYPE,
        ];
    }

    private function addSelectClauses(QueryBuilder $qb, DateTimeImmutable $calcDate): void
    {
        // add regular fields
        foreach ([
            'startDate',
            'endDate',
            'note',
            'createdAt',
            'updatedAt',
        ] as $field) {
            $qb->addSelect(sprintf('acpw.%s AS %s', $field, $field));
        }

        // those with identity
        foreach ([
            'createdBy',
            'updatedBy',
            'handlingThierParty',
        ] as $field) {
            $qb->addSelect(sprintf('IDENTITY(acpw.%s) AS %s', $field, $field));
        }

        // join socialaction
        $qb
            ->join('acpw.socialAction', 'sa')
            ->addSelect('sa.id AS socialActionId')
            ->addSelect('sa.id AS socialAction')
            ->addSelect('IDENTITY(sa.issue) AS socialIssue');

        // join acp
        $qb
            ->addSelect('acp.id AS acp_id')
            ->addSelect('IDENTITY(acp.user) AS acp_user');

        // persons
        $qb
            ->addSelect('(SELECT AGGREGATE(person_acpw_member.id) FROM ' . Person::class . ' person_acpw_member '
                . 'WHERE person_acpw_member MEMBER OF acpw.persons) AS personsId')
            ->addSelect('(SELECT AGGREGATE(person1_acpw_member.id) FROM ' . Person::class . ' person1_acpw_member '
                . 'WHERE person1_acpw_member MEMBER OF acpw.persons) AS personsName');

        // referrers => at date XXXX
        $qb
            ->addSelect('(SELECT IDENTITY(history.user) FROM ' . UserHistory::class . ' history ' .
            'WHERE history.accompanyingPeriod = acp AND history.startDate <= :calcDate AND (history.endDate IS NULL OR history.endDate > :calcDate)) AS referrers');

        // thirdparties
        $qb
            ->addSelect('(SELECT AGGREGATE(tp.id) FROM ' . ThirdParty::class . ' tp '
                . 'WHERE tp MEMBER OF acpw.thirdParties) AS thirdParties');

        // goals
        $qb
            ->addSelect('(SELECT AGGREGATE(IDENTITY(goal.goal)) FROM ' . AccompanyingPeriodWorkGoal::class . ' goal '
                . 'WHERE goal MEMBER OF acpw.goals) AS goalsId')
            ->addSelect('(SELECT AGGREGATE(g.title) FROM ' . AccompanyingPeriodWorkGoal::class . ' goal1 '
                . 'LEFT JOIN ' . Goal::class . ' g WITH goal1.goal = g.id WHERE goal1 MEMBER OF acpw.goals) AS goalsTitle');

        // goals results
        $qb
            ->addSelect('(SELECT AGGREGATE(wr.id) FROM ' . Result::class . ' wr '
                . 'JOIN ' . AccompanyingPeriodWorkGoal::class . ' wg WITH wr MEMBER OF wg.results '
                . 'WHERE wg MEMBER OF acpw.goals) AS goalResultsId')
            ->addSelect('(SELECT AGGREGATE(wr1.title) FROM ' . Result::class . ' wr1 '
                . 'JOIN ' . AccompanyingPeriodWorkGoal::class . ' wg1 WITH wr1 MEMBER OF wg1.results '
                . 'WHERE wg1 MEMBER OF acpw.goals) AS goalResultsTitle');

        // results
        $qb
            ->addSelect('(SELECT AGGREGATE(result.id) FROM ' . Result::class . ' result '
                . 'WHERE result MEMBER OF acpw.results ) AS resultsId ')
            ->addSelect('(SELECT AGGREGATE (result1.title) FROM ' . Result::class . ' result1 '
                . 'WHERE result1 MEMBER OF acpw.results ) AS resultsTitle ');

        // evaluations
        $qb
            ->addSelect('(SELECT AGGREGATE(IDENTITY(we.evaluation)) FROM ' . AccompanyingPeriodWorkEvaluation::class . ' we '
                . 'WHERE we MEMBER OF acpw.accompanyingPeriodWorkEvaluations ) AS evaluationsId ')
            ->addSelect('(SELECT AGGREGATE(ev.title) FROM ' . AccompanyingPeriodWorkEvaluation::class . ' we1 '
                . 'LEFT JOIN ' . Evaluation::class . ' ev WITH we1.evaluation = ev.id '
                . 'WHERE we1 MEMBER OF acpw.accompanyingPeriodWorkEvaluations ) AS evaluationsTitle ');

        $qb->setParameter('calcDate', $calcDate);
    }
}
