<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Export;

use Chill\MainBundle\Export\FormatterInterface;
use Chill\MainBundle\Export\GroupedExportInterface;
use Chill\MainBundle\Export\Helper\AggregateStringHelper;
use Chill\MainBundle\Export\Helper\DateTimeHelper;
use Chill\MainBundle\Export\Helper\TranslatableStringExportLabelHelper;
use Chill\MainBundle\Export\Helper\UserHelper;
use Chill\MainBundle\Export\ListInterface;
use Chill\MainBundle\Form\Type\PickRollingDateType;
//use Chill\MainBundle\Service\RollingDate\RollingDateConverter;
use Chill\MainBundle\Service\RollingDate\RollingDate;
use Chill\MainBundle\Service\RollingDate\RollingDateConverterInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod\AccompanyingPeriodWorkEvaluation;
use Chill\PersonBundle\Entity\AccompanyingPeriod\UserHistory;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Entity\Person\PersonCenterHistory;
use Chill\PersonBundle\Export\Declarations;
use Chill\PersonBundle\Export\Helper\LabelPersonHelper;
use Chill\PersonBundle\Repository\SocialWork\SocialActionRepository;
use Chill\PersonBundle\Repository\SocialWork\SocialIssueRepository;
use Chill\PersonBundle\Security\Authorization\AccompanyingPeriodVoter;
use Chill\PersonBundle\Templating\Entity\SocialActionRender;
use Chill\PersonBundle\Templating\Entity\SocialIssueRender;
use DateTimeImmutable;
use Doctrine\ORM\AbstractQuery;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormBuilderInterface;

class ListEvaluation implements ListInterface, GroupedExportInterface
{
    private const FIELDS = [
        'id',
        'startDate',
        'endDate',
        'maxDate',
        'warningInterval',
        'timeSpent',
        'acpw_id',
        'acpw_startDate',
        'acpw_endDate',
        'acpw_socialaction_id',
        'acpw_socialaction',
        'acpw_socialissue',
        'acpw_referrers',
        'acpw_note',
        'acpw_acp_id',
        'acpw_acp_user',
        'acpw_persons_id',
        'acpw_persons',
        'comment',
        'eval_title',
        'createdAt',
        'updatedAt',
        'createdBy',
        'updatedBy',
    ];

    private AggregateStringHelper $aggregateStringHelper;

    private DateTimeHelper $dateTimeHelper;

    private EntityManagerInterface $entityManager;

    private LabelPersonHelper $personHelper;

    private RollingDateConverterInterface $rollingDateConverter;

    private SocialActionRender $socialActionRender;

    private SocialActionRepository $socialActionRepository;

    private SocialIssueRender $socialIssueRender;

    private SocialIssueRepository $socialIssueRepository;

    private TranslatableStringExportLabelHelper $translatableStringExportLabelHelper;

    private UserHelper $userHelper;

    public function __construct(
        EntityManagerInterface $entityManager,
        SocialIssueRender $socialIssueRender,
        SocialIssueRepository $socialIssueRepository,
        SocialActionRender $socialActionRender,
        SocialActionRepository $socialActionRepository,
        UserHelper $userHelper,
        LabelPersonHelper $personHelper,
        DateTimeHelper $dateTimeHelper,
        TranslatableStringExportLabelHelper $translatableStringExportLabelHelper,
        AggregateStringHelper $aggregateStringHelper,
        RollingDateConverterInterface $rollingDateConverter
    ) {
        $this->entityManager = $entityManager;
        $this->socialIssueRender = $socialIssueRender;
        $this->socialIssueRepository = $socialIssueRepository;
        $this->socialActionRender = $socialActionRender;
        $this->socialActionRepository = $socialActionRepository;
        $this->userHelper = $userHelper;
        $this->personHelper = $personHelper;
        $this->dateTimeHelper = $dateTimeHelper;
        $this->translatableStringExportLabelHelper = $translatableStringExportLabelHelper;
        $this->aggregateStringHelper = $aggregateStringHelper;
        $this->rollingDateConverter = $rollingDateConverter;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder
            ->add('calc_date', PickRollingDateType::class, [
                'label' => 'export.list.eval.Date of calculation for associated elements',
                'help' => 'export.list.eval.help_description',
                'required' => true,
                'data' => new RollingDate(RollingDate::T_TODAY),
            ]);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_LIST];
    }

    public function getDescription(): string
    {
        return 'export.list.eval.Generate a list of evaluations, filtered on different parameters';
    }

    public function getGroup(): string
    {
        return 'Exports of evaluations';
    }

    public function getLabels($key, array $values, $data)
    {
        switch ($key) {
            case 'startDate':
            case 'endDate':
            case 'maxDate':
            case 'acpw_startDate':
            case 'acpw_endDate':
            case 'createdAt':
            case 'updatedAt':
                return $this->dateTimeHelper->getLabel('export.list.eval.' . $key);

            case 'acpw_socialaction':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.eval.' . $key;
                    }

                    if (null === $value || '' === $value) {
                        return '';
                    }

                    return $this->socialActionRender->renderString(
                        $this->socialActionRepository->find($value),
                        []
                    );
                };

            case 'acpw_socialissue':
                return function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.eval.' . $key;
                    }

                    if (null === $value || '' === $value) {
                        return '';
                    }

                    return $this->socialIssueRender->renderString(
                        $this->socialIssueRepository->find($value),
                        []
                    );
                };

            case 'createdBy':
            case 'updatedBy':
            case 'acpw_acp_user':
                return $this->userHelper->getLabel($key, $values, 'export.list.eval.' . $key);

            case 'acpw_referrers':
                return $this->userHelper->getLabel($key, $values, 'export.list.eval.' . $key);

            case 'acpw_persons_id':
                return $this->aggregateStringHelper->getLabelMulti($key, $values, 'export.list.eval.' . $key);

            case 'acpw_persons':
                return $this->personHelper->getLabelMulti($key, $values, 'export.list.eval.' . $key);

            case 'eval_title':
                return $this->translatableStringExportLabelHelper
                    ->getLabel($key, $values, 'export.list.eval.' . $key);

            default:
                return static function ($value) use ($key) {
                    if ('_header' === $value) {
                        return 'export.list.eval.' . $key;
                    }

                    if (null === $value) {
                        return '';
                    }

                    return $value;
                };
        }
    }

    public function getQueryKeys($data)
    {
        return self::FIELDS;
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(AbstractQuery::HYDRATE_SCALAR);
    }

    public function getTitle(): string
    {
        return 'export.list.eval.List of evaluations';
    }

    public function getType(): string
    {
        return Declarations::EVAL_TYPE;
    }

    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(static function ($el) {
            return $el['center'];
        }, $acl);

        $qb = $this->entityManager->createQueryBuilder();

        $qb
            ->from(AccompanyingPeriodWorkEvaluation::class, 'workeval')
            ->distinct()
            ->select('workeval.id AS id')
            ->join('workeval.accompanyingPeriodWork', 'acpw')
            ->join('acpw.accompanyingPeriod', 'acp')
            ->join('acp.participations', 'acppart')
            ->join('acppart.person', 'person')
            // ignore participation which didn't last one day, at least
            ->andWhere('acppart.startDate != acppart.endDate OR acppart.endDate IS NULL')
            // get participants at the given date
            ->andWhere('acppart.startDate <= :calc_date AND (acppart.endDate > :calc_date OR acppart.endDate IS NULL)')
            ->andWhere(
                $qb->expr()->exists(
                    'SELECT 1 FROM ' . PersonCenterHistory::class . ' acl_count_person_history WHERE acl_count_person_history.person = person
                        AND acl_count_person_history.center IN (:authorized_centers)
                        '
                )
            )
            ->setParameter('authorized_centers', $centers)
            ->setParameter('calc_date', $this->rollingDateConverter->convert($data['calc_date']));

        $this->addSelectClauses($qb, $this->rollingDateConverter->convert($data['calc_date']));

        return $qb;
    }

    public function requiredRole(): string
    {
        return AccompanyingPeriodVoter::STATS;
    }

    public function supportsModifiers(): array
    {
        return [
            Declarations::EVAL_TYPE,
            Declarations::SOCIAL_WORK_ACTION_TYPE,
            Declarations::ACP_TYPE,
            Declarations::PERSON_TYPE,
        ];
    }

    private function addSelectClauses(QueryBuilder $qb, DateTimeImmutable $calc_date): void
    {
        // add the regular fields
        foreach (['startDate', 'endDate', 'maxDate', 'warningInterval', 'comment', 'createdAt', 'updatedAt'] as $field) {
            $qb->addSelect(sprintf('workeval.%s AS %s', $field, $field));
        }

        // add the time spent field
        $qb->addSelect('(workeval.timeSpent / 60) AS timeSpent');

        // those with identity
        foreach (['createdBy', 'updatedBy'] as $field) {
            $qb->addSelect(sprintf('IDENTITY(workeval.%s) AS %s', $field, $field));
        }

        foreach (['id', 'startDate', 'endDate', 'note'] as $field) {
            $qb->addSelect(sprintf('acpw.%s AS %s', $field, 'acpw_' . $field));
        }

        // join socialaction
        $qb
            ->leftJoin('acpw.socialAction', 'sa')
            ->addSelect('sa.id AS acpw_socialaction_id')
            ->addSelect('sa.id AS acpw_socialaction')
            ->addSelect('IDENTITY(sa.issue) AS acpw_socialissue');

        // join acp
        $qb
            ->addSelect('acp.id AS acpw_acp_id')
            ->addSelect('IDENTITY(acp.user) AS acpw_acp_user');

        // referrers => at date XXXX
        $qb
            ->addSelect('(SELECT IDENTITY(history.user) FROM ' . UserHistory::class . ' history ' .
                'WHERE history.accompanyingPeriod = acp AND history.startDate <= :calc_date AND (history.endDate IS NULL OR history.endDate > :calc_date)) AS acpw_referrers');

        // persons
        $qb
            ->addSelect('(SELECT AGGREGATE(person_acpw_member.id) FROM ' . Person::class . ' person_acpw_member '
                . 'WHERE person_acpw_member MEMBER OF acpw.persons) AS acpw_persons_id')
            ->addSelect('(SELECT AGGREGATE(person1_acpw_member.id) FROM ' . Person::class . ' person1_acpw_member '
                . 'WHERE person1_acpw_member MEMBER OF acpw.persons) AS acpw_persons');

        // join evaluation
        $qb
            ->leftJoin('workeval.evaluation', 'eval')
            ->addSelect('eval.title AS eval_title');
    }
}
